/*
 * pop.c
 * This file is part of LibPOPc
 *
 * Copyright (C) 2022 - Félix Arreola Rodríguez
 *
 * LibPOPc is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * LibPOPc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LibPOPc; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <errno.h>
#include <ctype.h>

#include "pop.h"

#define POP_WANT_FLAGS_OR_POLL(var, flags, poll) \
	do { \
	if ((var & flags) == 0) { \
		poll = flags; \
		return POP_POLL; \
	} \
	} while (0)

enum {
	POP_LINE_INVALID = 0,
	POP_LINE_OK,
	POP_LINE_ERR
};

enum {
	POP_INTERNAL_FREE = 0,
	POP_INTERNAL_WRITE,
	POP_INTERNAL_READ,
};

struct _POPSession {
	/* El socket de datos */
	int sock;
	
	/* Estado de la sessión en general */
	int estado;
	
	/* Estado interno, lectura, escritura */
	int internal;
	
	/* Variable que controla el poller */
	int poll;
	
	/* La conexión con el servidor */
	struct sockaddr_storage server;
	
	/* El buffer de lectura de la red */
	char read_buffer[8192];
	size_t read_count;
	int ceof;
	
	/* Bandera que me indica si el último comando que ejecutamos esperamos multi-lineas */
	int waiting_multiline;
	
	char *user;
	char *passwd;
	int last_code;
	char last_cmd[8192];
	char last_message[1024];
	
	/* El stream que usaremos para recibir un correo */
	FILE *data_stream;
	
	/* Función para llamar cuando lea entre lineas */
	POPDataFunc line_func;
	void *line_func_data;
};

int _pop_session_internal_connect_cont (POPSession *session);
int _pop_session_internal_banner_code (POPSession *session);
int _pop_session_internal_send_helo (POPSession *session);
int _pop_session_internal_helo_code (POPSession *session);
int _pop_session_internal_start_auth_login (POPSession *session);
int _pop_session_internal_auth_login_code (POPSession *session);
int _pop_session_internal_send_user (POPSession *session);
int _pop_session_internal_user_code (POPSession *session);
int _pop_session_internal_send_pass (POPSession *session);
int _pop_session_internal_pass_code (POPSession *session);
int _pop_session_internal_send_cmd (POPSession *session);
int _pop_session_internal_cmd_code (POPSession *session);
int _pop_session_internal_send_close (POPSession *session);
int _pop_session_internal_close_code (POPSession *session);
int _pop_session_internal_send_data (POPSession *session);
int _pop_session_internal_data_code (POPSession *session);
int _pop_session_internal_data_write_finish (POPSession *session);
int _pop_session_internal_data_write (POPSession *session);

int pop_session_read_bytes (POPSession *session);
int pop_session_readline (POPSession *session, char *buffer, int max);
int _pop_session_validate_response_line (POPSession *session, char *buffer, int size);

POPSession *pop_session_create (void) {
	POPSession *pop;
	
	pop = (POPSession *) malloc (sizeof (POPSession));
	
	if (pop == NULL) return NULL;
	
	memset (pop, 0, sizeof (POPSession));
	
	pop->sock = -1;
	
	return pop;
}

void pop_session_free (POPSession *session) {
	pop_session_close_and_clean (session);
	
	free (session);
}

void pop_session_close_and_clean (POPSession *session) {
	if (session->user != NULL) {
		free (session->user);
		session->user = NULL;
	}
	
	if (session->passwd != NULL) {
		free (session->passwd);
		session->passwd = NULL;
	}
	
	if (session->sock >= 0) {
		close (session->sock);
		session->sock = -1;
	}
	
	session->internal = POP_INTERNAL_FREE;
	session->estado = POP_STATE_CLOSED;
}

// Arreglada
int pop_session_connect (POPSession *session, struct sockaddr *server, socklen_t addr_len) {
	int c;
	int res;
	
	session->poll = POP_WANT_NONE;
	
	c = socket (server->sa_family, SOCK_STREAM, 0);
	
	if (c < 0) {
		return POP_ERROR;
	}
	
	fcntl (c, F_SETFL, O_NONBLOCK);
	
	session->estado = POP_STATE_TRYING_CONNECT;
	memcpy (&session->server, server, addr_len);
	
	res = connect (c, server, addr_len);
	
	if (res < 0) {
		if (errno == EINPROGRESS) {
			session->sock = c;
			session->poll = POP_WANT_POLL_OUT;
			session->internal = POP_INTERNAL_WRITE;
			session->estado = POP_STATE_TRYING_CONNECT;
			return POP_POLL;
		} else {
			session->sock = -1;
			close (c);
			
			return POP_ERROR;
		}
	} else {
		/* Conexión exitosa */
		session->sock = c;
		session->internal = POP_INTERNAL_READ;
		session->estado = POP_STATE_BANNER;
		session->read_count = 0;
		return pop_session_continue (session, POP_WANT_POLL_IN);
	}
}

// FIXME:
int pop_session_continue (POPSession *session, int what) {
	char linea[8192];
	size_t linea_size;
	int res;
	int e;
	int use_short;
	
	if (session->internal == POP_INTERNAL_FREE) {
		return POP_ERROR;
	}
	
	if (session->internal == POP_INTERNAL_WRITE) {
		POP_WANT_FLAGS_OR_POLL (what, POP_WANT_POLL_OUT, session->poll);
		
		/* Se logró el evento esperado */
		switch (session->estado) {
			case POP_STATE_TRYING_CONNECT:
				return _pop_session_internal_connect_cont (session);
			case POP_STATE_AUTH_LOGIN_USER:
				return _pop_session_internal_send_user (session);
			case POP_STATE_AUTH_LOGIN_PASS:
				return _pop_session_internal_send_pass (session);
			case POP_STATE_CMD:
				return _pop_session_internal_send_cmd (session);
			// FIXME: Arreglar esto
			case POP_STATE_CLOSING:
				return _pop_session_internal_send_close (session);
		}
	} else if (session->internal == POP_INTERNAL_READ) {
		POP_WANT_FLAGS_OR_POLL (what, POP_WANT_POLL_IN, session->poll);

pop_read_a_line_again:
		/* Leer una linea */
		linea_size = pop_session_readline (session, linea, sizeof (linea));
		if (linea_size == POP_POLL) {
			/* Si al intentar leer, no hay linea, revisar si podemos leer algunos bytes */
			res = pop_session_read_bytes (session);
			if (res == POP_ERROR) {
				pop_session_close_and_clean (session);
				return POP_ERROR;
			} else if (res == POP_POLL) {
				/* No hay bytes, hacer polling */
				session->poll = POP_WANT_POLL_IN;
				return POP_POLL;
			} else if (session->ceof == 1) {
				/* Cierre de la conexión, limpiar y cerrar */
				if (session->estado == POP_STATE_CLOSING) {
					pop_session_close_and_clean (session);
					return POP_SUCCESS;
				} else {
					pop_session_close_and_clean (session);
					return POP_ERROR;
				}
			}
			/* Con los bytes restantes de la lectura previa, intentar leer la linea */
			linea_size = pop_session_readline (session, linea, sizeof (linea));
		}
		
		if (session->estado == POP_STATE_DATA) {
			/* Los datos no se validan como linea de respuesta */
			if ((linea[0] == '.' && linea[1] == 0) ||
			    (linea[0] == '.' && linea[1] == '\n')) {
				session->estado = POP_STATE_READY;
				session->internal = POP_INTERNAL_FREE;
				
				return POP_SUCCESS;
			}
			
			if (session->line_func != NULL) {
				session->line_func (session, linea, session->line_func_data);
			}
			goto pop_read_a_line_again;
		}
		
		res = _pop_session_validate_response_line (session, linea, linea_size);
		
		if (res == POP_LINE_INVALID) {
			/* Esperaba una respuesta +OK o -ERR */
			pop_session_close_and_clean (session);
			return POP_ERROR;
		}
		
		/* Linea buena, guardar el estado y el mensaje de la linea */
		if (res == POP_LINE_OK) {
			strncpy (session->last_message, &linea[4], sizeof (session->last_message) - 1);
		} else {
			strncpy (session->last_message, &linea[5], sizeof (session->last_message) - 1);
		}
		e = strlen (session->last_message) - 1;
		while (e > 0 && (session->last_message[e] == '\r' || session->last_message[e] == '\n')) {
			session->last_message[e] = 0;
			e--;
		}
		
		session->last_code = res;
		/* Correr la función correcta para pasar al siguiente estado */
		switch (session->estado) {
			case POP_STATE_BANNER:
				return _pop_session_internal_banner_code (session);
			case POP_STATE_AUTH_LOGIN_USER:
				return _pop_session_internal_user_code (session);
			case POP_STATE_AUTH_LOGIN_PASS:
				return _pop_session_internal_pass_code (session);
			case POP_STATE_CMD:
				return _pop_session_internal_cmd_code (session);
			// FIXME: Arreglar esto
			case POP_STATE_CLOSING:
				return _pop_session_internal_close_code (session);
		}
	}
}

// Arreglada
int _pop_session_validate_response_line (POPSession *session, char *buffer, int size) {
	int ok_or_err;
	
	if (size < 3) {
		return POP_LINE_INVALID;
	}
	
	ok_or_err = 0;
	/* Puede ser un +OK o un -ERR */
	if (strncmp ("+OK", buffer, 3) == 0) {
		ok_or_err = 1;
	} else if (size >= 4 && strncmp ("-ERR", buffer, 4) == 0) {
		ok_or_err = 2;
	}
	
	if (ok_or_err == 0) {
		return POP_LINE_INVALID;
	} else if (ok_or_err == 1) {
		return POP_LINE_OK;
	} else {
		return POP_LINE_ERR;
	}
}

int _pop_session_internal_connect_cont (POPSession *session) {
	int optval;
	socklen_t optlen;
	int res;
	
	optval = -1;
	
	optlen = sizeof (optval);
	
	res = getsockopt (session->sock, SOL_SOCKET, SO_ERROR, &optval, &optlen);

	if (res < 0) {
		pop_session_close_and_clean (session);

		return POP_ERROR;
	}
	
	if (optval != 0) {
		pop_session_close_and_clean (session);

		return POP_ERROR;
	}
	
	/* Pasamos a la lectura del banner */
	session->internal = POP_INTERNAL_READ;
	session->estado = POP_STATE_BANNER;
	session->read_count = 0;
	return pop_session_continue (session, POP_WANT_POLL_IN);
}

// Arreglada
int _pop_session_internal_banner_code (POPSession *session) {
	if (session->last_code == POP_LINE_OK) {
		session->internal = POP_INTERNAL_FREE;
		session->estado = POP_STATE_READY_BEFORE_LOGIN;
		return POP_SUCCESS;
	} else {
		pop_session_close_and_clean (session);
		return POP_ERROR;
	}
}

// Arreglada
int _pop_session_internal_start_auth_login (POPSession *session) {
	int res, linea_size;
	char *auth = "AUTH LOGIN\r\n";
	
	linea_size = strlen (auth);
	res = write (session->sock, auth, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = POP_WANT_POLL_OUT;
		session->internal = POP_INTERNAL_WRITE;
		return POP_POLL;
	} else if (res < linea_size) {
		pop_session_close_and_clean (session);
		
		return POP_ERROR;
	}
	
	/* Empezamos el AUTH LOGIN */
	session->internal = POP_INTERNAL_READ;
	
	return pop_session_continue (session, POP_WANT_POLL_IN);
}

// Arreglada
int _pop_session_internal_auth_login_code (POPSession *session) {
	if (session->last_code == POP_LINE_OK) {
		session->estado = POP_STATE_AUTH_LOGIN_USER;
		session->internal = POP_INTERNAL_WRITE;
		return pop_session_continue (session, POP_WANT_POLL_OUT);
	} else if (session->last_code == POP_LINE_ERR) {
		session->estado = POP_STATE_READY_BEFORE_LOGIN;
		session->internal = POP_INTERNAL_FREE;
		return POP_ERROR;
	}
}

// Arreglada
int _pop_session_internal_send_user (POPSession *session) {
	char linea[1024];
	int res, linea_size;
	
	linea_size = snprintf (linea, sizeof (linea), "USER %s\r\n", session->user);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = POP_WANT_POLL_OUT;
		session->internal = POP_INTERNAL_WRITE;
		return POP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		pop_session_close_and_clean (session);
		
		return POP_ERROR;
	}
	/* Envio completo */
	session->internal = POP_INTERNAL_READ;
	
	return pop_session_continue (session, POP_WANT_POLL_IN);
}

// Arreglada
int _pop_session_internal_user_code (POPSession *session) {
	if (session->last_code == POP_LINE_OK) {
		session->estado = POP_STATE_AUTH_LOGIN_PASS;
		session->internal = POP_INTERNAL_WRITE;
		return pop_session_continue (session, POP_WANT_POLL_OUT);
	} else if (session->last_code == POP_LINE_ERR) {
		session->internal = POP_INTERNAL_FREE;
		session->estado = POP_STATE_READY_BEFORE_LOGIN;
		
		return POP_ERROR;
	}
}

// Arreglada
int _pop_session_internal_send_pass (POPSession *session) {
	char linea[1024];
	int res, linea_size;
	
	linea_size = snprintf (linea, sizeof (linea), "PASS %s\r\n", session->passwd);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = POP_WANT_POLL_OUT;
		session->internal = POP_INTERNAL_WRITE;
		return POP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		pop_session_close_and_clean (session);
		
		return POP_ERROR;
	}
	/* Envio completo */
	session->internal = POP_INTERNAL_READ;
	
	return pop_session_continue (session, POP_WANT_POLL_IN);
}

// Arreglada
int _pop_session_internal_pass_code (POPSession *session) {
	if (session->last_code == POP_LINE_OK) {
		session->estado = POP_STATE_READY;
		session->internal = POP_INTERNAL_FREE;
		
		return POP_SUCCESS;
	} else if (session->last_code == POP_LINE_ERR) {
		session->estado = POP_STATE_READY_BEFORE_LOGIN;
		session->internal = POP_INTERNAL_FREE;
		
		return POP_ERROR;
	}
}

// Arreglada
int _pop_session_internal_send_cmd (POPSession *session) {
	int res, linea_size;
	
	linea_size = strlen (session->last_cmd);
	res = write (session->sock, session->last_cmd, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = POP_WANT_POLL_OUT;
		session->internal = POP_INTERNAL_WRITE;
		return POP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		pop_session_close_and_clean (session);
		
		return POP_ERROR;
	}
	/* Envio completo */
	session->internal = POP_INTERNAL_READ;
	
	return pop_session_continue (session, POP_WANT_POLL_IN);
}

// Arreglada
int _pop_session_internal_cmd_code (POPSession *session) {
	if (session->waiting_multiline && session->last_code == POP_LINE_OK) {
		/* Si estabamos esperando una rspuesta multilinea y obtuvimos +OK, pasar a otro estado */
		session->estado = POP_STATE_DATA;
		session->poll = POP_WANT_POLL_IN;
		session->internal = POP_INTERNAL_READ;
		
		return pop_session_continue (session, POP_WANT_POLL_IN);
	}
	
	/* En caso contrario, el comando ya terminó de ejecutarse, estamos listos para otro */
	session->estado = POP_STATE_READY;
	session->internal = POP_INTERNAL_FREE;
	
	if (session->last_code == POP_LINE_OK) {
		return POP_SUCCESS;
	}
	return POP_ERROR;
}

int _pop_session_internal_send_close (POPSession *session) {
	char *linea;
	int res, linea_size;
	
	linea = "QUIT\r\n";
	linea_size = strlen (linea);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = POP_WANT_POLL_OUT;
		session->internal = POP_INTERNAL_WRITE;
		return POP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		pop_session_close_and_clean (session);
		
		return POP_ERROR;
	}
	/* Envio completo */
	session->internal = POP_INTERNAL_READ;
	
	return pop_session_continue (session, POP_WANT_POLL_IN);
}

int _pop_session_internal_close_code (POPSession *session) {
	/* Realmente no importa si en el quit mandan una respuesta +OK o -Err, ya vamos de salida */
	session->internal = POP_INTERNAL_FREE;
	session->estado = POP_STATE_READY;
	pop_session_close_and_clean (session);
	
	return POP_SUCCESS;
}

int pop_session_read_bytes (POPSession *session) {
	int res;
	if (session->read_count >= sizeof (session->read_buffer)) {
		return POP_ERROR;
	}
	
	/* Leer tantos bytes como tenga el buffer */
	res = read (session->sock, &session->read_buffer[session->read_count], sizeof (session->read_buffer) - session->read_count);
	
	if (res == 0) {
		session->ceof = 1;
	} else if (res < 0) {
		if (errno == EINTR || errno == EAGAIN) {
			return POP_POLL;
		}
		return POP_ERROR;
	} else {
		session->read_count += res;
	}
	
	return POP_SUCCESS;
}

int pop_session_readline (POPSession *session, char *buffer, int max) {
	char *end;
	int line_size = 0;
	
	if (session->read_count > 0) {
		/* Ya había otros bytes en el buffer, procesarlos para buscar el salto de linea aquí */
		line_size = (max >= session->read_count) ? session->read_count : max - 1;
		
		end = memccpy (buffer, session->read_buffer, '\n', line_size);
		
		if (end != NULL) {
			line_size = end - buffer;
		}
		
		buffer[line_size] = 0;
		
		/* Consumir x bytes de la entrada */
		if (line_size < session->read_count) {
			memmove (session->read_buffer, &session->read_buffer[line_size], session->read_count - line_size);
			
			session->read_count -= line_size;
		} else {
			session->read_buffer[0] = 0;
			session->read_count = 0;
		}
		
		if (end != NULL) {
			
			if (buffer[line_size - 2] == '\r') {
				/* Quitar el \r */
				buffer[line_size - 2] = '\n';
				buffer[line_size - 1] = 0;
				line_size--;
			}
			
			return line_size;
		}
	}
	
	return POP_POLL;	
}

int pop_session_get_state (POPSession *session) {
	return session->estado;
}

int pop_session_auth_login (POPSession *session, const char *user, const char *passwd) {
	if (session->estado != POP_STATE_READY_BEFORE_LOGIN || session->internal != POP_INTERNAL_FREE) {
		return POP_ERROR;
	}
	
	/* Copiar el usuario y contraseña */
	session->user = strdup (user);
	if (passwd != NULL) {
		session->passwd = strdup (passwd);
	} else {
		session->passwd = NULL;
	}
	
	session->estado = POP_STATE_AUTH_LOGIN_USER;
	session->internal = POP_INTERNAL_WRITE;
	return pop_session_continue (session, POP_WANT_POLL_OUT);
}

// Arreglada
int pop_session_send_command (POPSession *session, char *command, int multiline) {
	int len;
	
	/* TODO: Revisar aquí otros estados */
	if (session->estado != POP_STATE_READY || session->internal != POP_INTERNAL_FREE) {
		return POP_ERROR;
	}
	
	/* Sanitizar el comando y asegurarnos que contenga el \r\n */
	len = strlen (command);
	if (command[len - 2] == '\r' && command[len - 1] == '\n') {
		if (len + 1 > sizeof (session->last_cmd)) {
			return POP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
	} else if (command[len - 1] == '\r') {
		if (len + 2 > sizeof (session->last_cmd)) {
			return POP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
		session->last_cmd[len] = '\n';
		session->last_cmd[len + 1] = 0;
	} else {
		if (len + 3 > sizeof (session->last_cmd)) {
			return POP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
		session->last_cmd[len] = '\r';
		session->last_cmd[len + 1] = '\n';
		session->last_cmd[len + 2] = 0;
	}
	
	session->estado = POP_STATE_CMD;
	session->internal = POP_INTERNAL_WRITE;
	session->waiting_multiline = multiline;
	
	return pop_session_continue (session, POP_WANT_POLL_OUT);
}

int pop_session_send_close (POPSession *session) {
	char *command;
	int len;
	
	if ((session->estado != POP_STATE_READY_BEFORE_LOGIN && session->estado != POP_STATE_READY) || session->internal != POP_INTERNAL_FREE) {
		return POP_ERROR;
	}
	
	session->estado = POP_STATE_CLOSING;
	session->internal = POP_INTERNAL_WRITE;
	
	return pop_session_continue (session, POP_WANT_POLL_OUT);
}

int pop_session_get_socket (POPSession *session) {
	return session->sock;
}

int pop_session_get_last_code (POPSession *session) {
	return session->last_code;
}

char * pop_session_get_last_message (POPSession *session) {
	return session->last_message;
}

void pop_session_data_func (POPSession *session, POPDataFunc func, void *data) {
	session->line_func = func;
	session->line_func_data = data;
}

int pop_session_get_poll (POPSession *session) {
	return session->poll;
}

